@@ -16,12 +16,17 @@ from __future__ import print_function
 
 import argparse
 import difflib
+import fnmatch
 import glob
 import os
 import shutil
 import subprocess
 import sys
 
+# The C++ standard whose features are required to build Binaryen.
+# Keep in sync with CMakeLists.txt CXX_STANDARD
+cxx_standard = 14
+
 
 def parse_args(args):
     usage_str = ("usage: 'python check.py [options]'\n\n"
@@ -81,6 +86,10 @@ def parse_args(args):
     parser.add_argument(
         '--list-suites', action='store_true',
         help='List the test suites that can be run.')
+    parser.add_argument(
+        '--filter', dest='test_name_filter', default='',
+        help=('Specifies a filter. Only tests whose paths contains this '
+              'substring will be run'))
 
     return parser.parse_args(args)
 
@@ -146,7 +155,13 @@ def which(program):
         if is_exe(program):
             return program
     else:
-        for path in os.environ["PATH"].split(os.pathsep):
+        paths = [
+            # Prefer tools installed using third_party/setup.py
+            os.path.join(options.binaryen_root, 'third_party', 'mozjs'),
+            os.path.join(options.binaryen_root, 'third_party', 'v8'),
+            os.path.join(options.binaryen_root, 'third_party', 'wabt', 'bin')
+        ] + os.environ['PATH'].split(os.pathsep)
+        for path in paths:
             path = path.strip('"')
             exe_file = os.path.join(path, program)
             if is_exe(exe_file):
@@ -170,7 +185,6 @@ NATIVEXX = (os.environ.get('CXX') or which('mingw32-g++') or
 NODEJS = os.getenv('NODE', which('nodejs') or which('node'))
 MOZJS = which('mozjs') or which('spidermonkey')
 V8 = which('v8') or which('d8')
-EMCC = which('emcc')
 
 BINARYEN_INSTALL_DIR = os.path.dirname(options.binaryen_bin)
 WASM_OPT = [os.path.join(options.binaryen_bin, 'wasm-opt')]
@@ -225,20 +239,16 @@ def has_shell_timeout():
 
 
 # Default options to pass to v8. These enable all features.
+# See https://github.com/v8/v8/blob/master/src/wasm/wasm-feature-flags.h
 V8_OPTS = [
+    '--wasm-staging',
     '--experimental-wasm-eh',
-    '--experimental-wasm-mv',
-    '--experimental-wasm-sat-f2i-conversions',
-    '--experimental-wasm-se',
-    '--experimental-wasm-threads',
     '--experimental-wasm-simd',
-    '--experimental-wasm-anyref',
-    '--experimental-wasm-bulk-memory',
+    '--experimental-wasm-reftypes',
+    '--experimental-wasm-compilation-hints',
     '--experimental-wasm-return-call'
 ]
 
-has_vanilla_llvm = False
-
 # external tools
 
 try:
@@ -262,26 +272,6 @@ if MOZJS is None:
     warn('no mozjs found (did not check native wasm support nor asm.js'
          ' validation)')
 
-try:
-    if EMCC is not None:
-        subprocess.check_call([EMCC, '--version'],
-                              stdout=subprocess.PIPE,
-                              stderr=subprocess.PIPE)
-except (OSError, subprocess.CalledProcessError):
-    EMCC = None
-if EMCC is None:
-    warn('no emcc found (did not check non-vanilla emscripten/binaryen'
-         ' integration)')
-
-has_vanilla_emcc = False
-try:
-    subprocess.check_call(
-        [os.path.join(options.binaryen_test, 'emscripten', 'emcc'), '--version'],
-        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    has_vanilla_emcc = True
-except (OSError, subprocess.CalledProcessError):
-    pass
-
 
 # utilities
 
@@ -370,7 +360,7 @@ def fail_if_not_contained(actual, expected):
 def fail_if_not_identical_to_file(actual, expected_file):
     binary = expected_file.endswith(".wasm") or type(actual) == bytes
     with open(expected_file, 'rb' if binary else 'r') as f:
-        fail_if_not_identical(f.read(), actual, fromfile=expected_file)
+        fail_if_not_identical(actual, f.read(), fromfile=expected_file)
 
 
 def get_test_dir(name):
@@ -387,15 +377,14 @@ def get_tests(test_dir, extensions=[]):
         tests += glob.glob(os.path.join(test_dir, '*'))
     for ext in extensions:
         tests += glob.glob(os.path.join(test_dir, '*' + ext))
+    if options.test_name_filter:
+        tests = fnmatch.filter(tests, options.test_name_filter)
     return sorted(tests)
 
 
 if not options.interpreter:
     warn('no interpreter provided (did not test spec interpreter validation)')
 
-if not has_vanilla_emcc:
-    warn('no functional emcc submodule found')
-
 
 if not options.spec_tests:
     options.spec_tests = get_tests(get_test_dir('spec'), ['.wast'])
@@ -404,13 +393,13 @@ else:
 
 # 11/27/2019: We updated the spec test suite to upstream spec repo. For some
 # files that started failing after this update, we added the new files to this
-# blacklist and preserved old ones by renaming them to 'old_[FILENAME].wast'
+# skip-list and preserved old ones by renaming them to 'old_[FILENAME].wast'
 # not to lose coverage. When the cause of the error is fixed or the unsupported
 # construct gets support so the new test passes, we can delete the
 # corresponding 'old_[FILENAME].wast' file. When you fix the new file and
 # delete the old file, make sure you rename the corresponding .wast.log file in
 # expected-output/ if any.
-SPEC_TEST_BLACKLIST = [
+SPEC_TESTS_TO_SKIP = [
     # Stacky code / notation
     'block.wast',
     'call.wast',
@@ -465,21 +454,11 @@ SPEC_TEST_BLACKLIST = [
     'unreached-invalid.wast'  # 'assert_invalid' failure
 ]
 options.spec_tests = [t for t in options.spec_tests if os.path.basename(t) not
-                      in SPEC_TEST_BLACKLIST]
+                      in SPEC_TESTS_TO_SKIP]
 
 
 # check utilities
 
-
-def validate_binary(wasm):
-    if V8:
-        cmd = [V8] + V8_OPTS + [in_binaryen('scripts', 'validation_shell.js'), '--', wasm]
-        print('            ', ' '.join(cmd))
-        subprocess.check_call(cmd, stdout=subprocess.PIPE)
-    else:
-        print('(skipping v8 binary validation)')
-
-
 def binary_format_check(wast, verify_final_result=True, wasm_as_args=['-g'],
                         binary_suffix='.fromBinary', original_wast=None):
     # checks we can convert the wast to binary and back
@@ -492,13 +471,6 @@ def binary_format_check(wast, verify_final_result=True, wasm_as_args=['-g'],
     subprocess.check_call(cmd, stdout=subprocess.PIPE)
     assert os.path.exists('a.wasm')
 
-    # make sure it is a valid wasm, using a real wasm VM
-    if os.path.basename(original_wast or wast) not in [
-        'atomics.wast',    # https://bugs.chromium.org/p/v8/issues/detail?id=9425
-        'simd.wast',    # https://bugs.chromium.org/p/v8/issues/detail?id=8460
-    ]:
-        validate_binary('a.wasm')
-
     cmd = WASM_DIS + ['a.wasm', '-o', 'ab.wast']
     print('            ', ' '.join(cmd))
     if os.path.exists('ab.wast'):
@@ -552,3 +524,13 @@ def with_pass_debug(check):
         else:
             if 'BINARYEN_PASS_DEBUG' in os.environ:
                 del os.environ['BINARYEN_PASS_DEBUG']
+
+
+# checks if we are on windows, and if so logs out that a test is being skipped,
+# and returns True. This is a central location for all test skipping on
+# windows, so that we can easily find which tests are skipped.
+def skip_if_on_windows(name):
+    if get_platform() == 'windows':
+        print('skipping test "%s" on windows' % name)
+        return True
+    return False
