@@ -78,7 +78,7 @@ struct ExpressionAnalyzer {
 
   // hash an expression, ignoring superficial details like specific internal
   // names
-  static HashType hash(Expression* curr);
+  static size_t hash(Expression* curr);
 };
 
 // Re-Finalizes all node types. This can be run after code was modified in
@@ -145,18 +145,35 @@ struct ReFinalize
   void visitSelect(Select* curr);
   void visitDrop(Drop* curr);
   void visitReturn(Return* curr);
-  void visitHost(Host* curr);
+  void visitMemorySize(MemorySize* curr);
+  void visitMemoryGrow(MemoryGrow* curr);
   void visitRefNull(RefNull* curr);
   void visitRefIsNull(RefIsNull* curr);
   void visitRefFunc(RefFunc* curr);
+  void visitRefEq(RefEq* curr);
   void visitTry(Try* curr);
   void visitThrow(Throw* curr);
   void visitRethrow(Rethrow* curr);
   void visitBrOnExn(BrOnExn* curr);
   void visitNop(Nop* curr);
   void visitUnreachable(Unreachable* curr);
-  void visitPush(Push* curr);
   void visitPop(Pop* curr);
+  void visitTupleMake(TupleMake* curr);
+  void visitTupleExtract(TupleExtract* curr);
+  void visitI31New(I31New* curr);
+  void visitI31Get(I31Get* curr);
+  void visitRefTest(RefTest* curr);
+  void visitRefCast(RefCast* curr);
+  void visitBrOnCast(BrOnCast* curr);
+  void visitRttCanon(RttCanon* curr);
+  void visitRttSub(RttSub* curr);
+  void visitStructNew(StructNew* curr);
+  void visitStructGet(StructGet* curr);
+  void visitStructSet(StructSet* curr);
+  void visitArrayNew(ArrayNew* curr);
+  void visitArrayGet(ArrayGet* curr);
+  void visitArraySet(ArraySet* curr);
+  void visitArrayLen(ArrayLen* curr);
 
   void visitFunction(Function* curr);
 
@@ -212,18 +229,35 @@ struct ReFinalizeNode : public OverriddenVisitor<ReFinalizeNode> {
   void visitSelect(Select* curr) { curr->finalize(); }
   void visitDrop(Drop* curr) { curr->finalize(); }
   void visitReturn(Return* curr) { curr->finalize(); }
-  void visitHost(Host* curr) { curr->finalize(); }
+  void visitMemorySize(MemorySize* curr) { curr->finalize(); }
+  void visitMemoryGrow(MemoryGrow* curr) { curr->finalize(); }
   void visitRefNull(RefNull* curr) { curr->finalize(); }
   void visitRefIsNull(RefIsNull* curr) { curr->finalize(); }
   void visitRefFunc(RefFunc* curr) { curr->finalize(); }
+  void visitRefEq(RefEq* curr) { curr->finalize(); }
   void visitTry(Try* curr) { curr->finalize(); }
   void visitThrow(Throw* curr) { curr->finalize(); }
   void visitRethrow(Rethrow* curr) { curr->finalize(); }
   void visitBrOnExn(BrOnExn* curr) { curr->finalize(); }
   void visitNop(Nop* curr) { curr->finalize(); }
   void visitUnreachable(Unreachable* curr) { curr->finalize(); }
-  void visitPush(Push* curr) { curr->finalize(); }
   void visitPop(Pop* curr) { curr->finalize(); }
+  void visitTupleMake(TupleMake* curr) { curr->finalize(); }
+  void visitTupleExtract(TupleExtract* curr) { curr->finalize(); }
+  void visitI31New(I31New* curr) { curr->finalize(); }
+  void visitI31Get(I31Get* curr) { curr->finalize(); }
+  void visitRefTest(RefTest* curr) { curr->finalize(); }
+  void visitRefCast(RefCast* curr) { curr->finalize(); }
+  void visitBrOnCast(BrOnCast* curr) { curr->finalize(); }
+  void visitRttCanon(RttCanon* curr) { curr->finalize(); }
+  void visitRttSub(RttSub* curr) { curr->finalize(); }
+  void visitStructNew(StructNew* curr) { curr->finalize(); }
+  void visitStructGet(StructGet* curr) { curr->finalize(); }
+  void visitStructSet(StructSet* curr) { curr->finalize(); }
+  void visitArrayNew(ArrayNew* curr) { curr->finalize(); }
+  void visitArrayGet(ArrayGet* curr) { curr->finalize(); }
+  void visitArraySet(ArraySet* curr) { curr->finalize(); }
+  void visitArrayLen(ArrayLen* curr) { curr->finalize(); }
 
   void visitExport(Export* curr) { WASM_UNREACHABLE("unimp"); }
   void visitGlobal(Global* curr) { WASM_UNREACHABLE("unimp"); }
@@ -301,6 +335,20 @@ struct AutoDrop : public WalkerPass<ExpressionStackWalker<AutoDrop>> {
     }
   }
 
+  void visitTry(Try* curr) {
+    bool acted = false;
+    if (maybeDrop(curr->body)) {
+      acted = true;
+    }
+    if (maybeDrop(curr->catchBody)) {
+      acted = true;
+    }
+    if (acted) {
+      reFinalize();
+      assert(curr->type == Type::none);
+    }
+  }
+
   void doWalkFunction(Function* curr) {
     ReFinalize().walkFunctionInModule(curr, getModule());
     walk(curr->body);
@@ -319,7 +367,7 @@ struct I64Utilities {
       builder.makeUnary(ExtendUInt32, low),
       builder.makeBinary(ShlInt64,
                          builder.makeUnary(ExtendUInt32, high),
-                         builder.makeConst(Literal(int64_t(32)))));
+                         builder.makeConst(int64_t(32))));
   };
 
   static Expression* recreateI64(Builder& builder, Index low, Index high) {
@@ -333,7 +381,7 @@ struct I64Utilities {
       WrapInt64,
       builder.makeBinary(ShrUInt64,
                          builder.makeLocalGet(index, Type::i64),
-                         builder.makeConst(Literal(int64_t(32)))));
+                         builder.makeConst(int64_t(32))));
   }
 
   static Expression* getI64Low(Builder& builder, Index index) {
