@@ -18,13 +18,14 @@
 // Flattens code into "Flat IR" form. See ir/flat.h.
 //
 
-#include <wasm.h>
-#include <pass.h>
-#include <wasm-builder.h>
 #include <ir/branch-utils.h>
 #include <ir/effects.h>
 #include <ir/flat.h>
+#include <ir/properties.h>
 #include <ir/utils.h>
+#include <pass.h>
+#include <wasm-builder.h>
+#include <wasm.h>
 
 namespace wasm {
 
@@ -43,12 +44,15 @@ namespace wasm {
 // Once exception is that we allow an (unreachable) node, which is used
 // when we move something unreachable to another place, and need a
 // placeholder. We will never reach that (unreachable) anyhow
-struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpressionVisitor<Flatten>>> {
+struct Flatten
+  : public WalkerPass<
+      ExpressionStackWalker<Flatten, UnifiedExpressionVisitor<Flatten>>> {
   bool isFunctionParallel() override { return true; }
 
   Pass* create() override { return new Flatten; }
 
-  // For each expression, a bunch of expressions that should execute right before it
+  // For each expression, a bunch of expressions that should execute right
+  // before it
   std::unordered_map<Expression*, std::vector<Expression*>> preludes;
 
   // Break values are sent through a temp local
@@ -58,10 +62,23 @@ struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpress
     std::vector<Expression*> ourPreludes;
     Builder builder(*getModule());
 
-    if (Flat::isControlFlowStructure(curr)) {
+    // Nothing to do for constants and nop.
+    if (Properties::isConstantExpression(curr) || curr->is<Nop>()) {
+      return;
+    }
+
+    if (curr->is<Try>() || curr->is<Throw>() || curr->is<Rethrow>() ||
+        curr->is<BrOnExn>()) {
+      Fatal() << "Flatten does not support EH instructions yet";
+    }
+
+    if (Properties::isControlFlowStructure(curr)) {
       // handle control flow explicitly. our children do not have control flow,
       // but they do have preludes which we need to set up in the right place
-      assert(preludes.find(curr) == preludes.end()); // no one should have given us preludes, they are on the children
+
+      // no one should have given us preludes, they are on the children
+      assert(preludes.find(curr) == preludes.end());
+
       if (auto* block = curr->dynCast<Block>()) {
         // make a new list, where each item's preludes are added before it
         ExpressionList newList(getModule()->allocator);
@@ -79,7 +96,7 @@ struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpress
         block->list.swap(newList);
         // remove a block return value
         auto type = block->type;
-        if (isConcreteType(type)) {
+        if (type.isConcrete()) {
           // if there is a temp index for breaking to the block, use that
           Index temp;
           auto iter = breakTemps.find(block->name);
@@ -89,18 +106,19 @@ struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpress
             temp = builder.addVar(getFunction(), type);
           }
           auto*& last = block->list.back();
-          if (isConcreteType(last->type)) {
-            last = builder.makeSetLocal(temp, last);
+          if (last->type.isConcrete()) {
+            last = builder.makeLocalSet(temp, last);
           }
-          block->finalize(none);
+          block->finalize(Type::none);
           // and we leave just a get of the value
-          auto* rep = builder.makeGetLocal(temp, type);
+          auto* rep = builder.makeLocalGet(temp, type);
           replaceCurrent(rep);
           // the whole block is now a prelude
           ourPreludes.push_back(block);
         }
         // the block now has no return value, and may have become unreachable
-        block->finalize(none);
+        block->finalize(Type::none);
+
       } else if (auto* iff = curr->dynCast<If>()) {
         // condition preludes go before the entire if
         auto* rep = getPreludesWithExpression(iff->condition, iff);
@@ -109,142 +127,175 @@ struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpress
         auto* originalIfFalse = iff->ifFalse;
         auto type = iff->type;
         Expression* prelude = nullptr;
-        if (isConcreteType(type)) {
+        if (type.isConcrete()) {
           Index temp = builder.addVar(getFunction(), type);
-          if (isConcreteType(iff->ifTrue->type)) {
-            iff->ifTrue = builder.makeSetLocal(temp, iff->ifTrue);
+          if (iff->ifTrue->type.isConcrete()) {
+            iff->ifTrue = builder.makeLocalSet(temp, iff->ifTrue);
           }
-          if (iff->ifFalse && isConcreteType(iff->ifFalse->type)) {
-            iff->ifFalse = builder.makeSetLocal(temp, iff->ifFalse);
+          if (iff->ifFalse && iff->ifFalse->type.isConcrete()) {
+            iff->ifFalse = builder.makeLocalSet(temp, iff->ifFalse);
           }
           // the whole if (+any preludes from the condition) is now a prelude
           prelude = rep;
           // and we leave just a get of the value
-          rep = builder.makeGetLocal(temp, type);
+          rep = builder.makeLocalGet(temp, type);
         }
         iff->ifTrue = getPreludesWithExpression(originalIfTrue, iff->ifTrue);
-        if (iff->ifFalse) iff->ifFalse = getPreludesWithExpression(originalIfFalse, iff->ifFalse);
+        if (iff->ifFalse) {
+          iff->ifFalse =
+            getPreludesWithExpression(originalIfFalse, iff->ifFalse);
+        }
         iff->finalize();
         if (prelude) {
           ReFinalizeNode().visit(prelude);
           ourPreludes.push_back(prelude);
         }
         replaceCurrent(rep);
+
       } else if (auto* loop = curr->dynCast<Loop>()) {
         // remove a loop value
         Expression* rep = loop;
         auto* originalBody = loop->body;
         auto type = loop->type;
-        if (isConcreteType(type)) {
+        if (type.isConcrete()) {
           Index temp = builder.addVar(getFunction(), type);
-          loop->body = builder.makeSetLocal(temp, loop->body);
+          loop->body = builder.makeLocalSet(temp, loop->body);
           // and we leave just a get of the value
-          rep = builder.makeGetLocal(temp, type);
+          rep = builder.makeLocalGet(temp, type);
           // the whole if is now a prelude
           ourPreludes.push_back(loop);
-          loop->type = none;
+          loop->type = Type::none;
         }
         loop->body = getPreludesWithExpression(originalBody, loop->body);
         loop->finalize();
         replaceCurrent(rep);
+
       } else {
-        WASM_UNREACHABLE();
+        WASM_UNREACHABLE("unexpected expr type");
       }
+
     } else {
       // for anything else, there may be existing preludes
       auto iter = preludes.find(curr);
       if (iter != preludes.end()) {
         ourPreludes.swap(iter->second);
       }
+
       // special handling
-      if (auto* set = curr->dynCast<SetLocal>()) {
+      if (auto* set = curr->dynCast<LocalSet>()) {
         if (set->isTee()) {
           // we disallow local.tee
-          if (set->value->type == unreachable) {
+          if (set->value->type == Type::unreachable) {
             replaceCurrent(set->value); // trivial, no set happens
           } else {
             // use a set in a prelude + a get
-            set->setTee(false);
+            set->makeSet();
             ourPreludes.push_back(set);
-            replaceCurrent(builder.makeGetLocal(set->index, set->value->type));
+            Type localType = getFunction()->getLocalType(set->index);
+            replaceCurrent(builder.makeLocalGet(set->index, localType));
           }
         }
+
       } else if (auto* br = curr->dynCast<Break>()) {
         if (br->value) {
           auto type = br->value->type;
-          if (isConcreteType(type)) {
+          if (type.isConcrete()) {
             // we are sending a value. use a local instead
-            Index temp = getTempForBreakTarget(br->name, type);
-            ourPreludes.push_back(builder.makeSetLocal(temp, br->value));
+            Type blockType = findBreakTarget(br->name)->type;
+            Index temp = getTempForBreakTarget(br->name, blockType);
+            ourPreludes.push_back(builder.makeLocalSet(temp, br->value));
+
+            // br_if leaves a value on the stack if not taken, which later can
+            // be the last element of the enclosing innermost block and flow
+            // out. The local we created using 'getTempForBreakTarget' returns
+            // the return type of the block this branch is targetting, which may
+            // not be the same with the innermost block's return type. For
+            // example,
+            // (block $any (result anyref)
+            //   (block (result funcref)
+            //     (local.tee $0
+            //       (br_if $any
+            //         (ref.null func)
+            //         (i32.const 0)
+            //       )
+            //     )
+            //   )
+            // )
+            // In this case we need two locals to store (ref.null); one with
+            // funcref type that's for the target block ($label0) and one more
+            // with anyref type in case for flowing out. Here we create the
+            // second 'flowing out' local in case two block's types are
+            // different.
+            if (type != blockType) {
+              temp = builder.addVar(getFunction(), type);
+              ourPreludes.push_back(builder.makeLocalSet(
+                temp, ExpressionManipulator::copy(br->value, *getModule())));
+            }
+
             if (br->condition) {
               // the value must also flow out
               ourPreludes.push_back(br);
-              if (isConcreteType(br->type)) {
-                replaceCurrent(builder.makeGetLocal(temp, type));
+              if (br->type.isConcrete()) {
+                replaceCurrent(builder.makeLocalGet(temp, type));
               } else {
-                assert(br->type == unreachable);
+                assert(br->type == Type::unreachable);
                 replaceCurrent(builder.makeUnreachable());
               }
             }
             br->value = nullptr;
             br->finalize();
           } else {
-            assert(type == unreachable);
+            assert(type == Type::unreachable);
             // we don't need the br at all
             replaceCurrent(br->value);
           }
         }
+
       } else if (auto* sw = curr->dynCast<Switch>()) {
         if (sw->value) {
           auto type = sw->value->type;
-          if (isConcreteType(type)) {
+          if (type.isConcrete()) {
             // we are sending a value. use a local instead
             Index temp = builder.addVar(getFunction(), type);
-            ourPreludes.push_back(builder.makeSetLocal(temp, sw->value));
+            ourPreludes.push_back(builder.makeLocalSet(temp, sw->value));
             // we don't know which break target will be hit - assign to them all
             auto names = BranchUtils::getUniqueTargets(sw);
             for (auto name : names) {
-              ourPreludes.push_back(builder.makeSetLocal(
-                getTempForBreakTarget(name, type),
-                builder.makeGetLocal(temp, type)
-              ));
+              ourPreludes.push_back(
+                builder.makeLocalSet(getTempForBreakTarget(name, type),
+                                     builder.makeLocalGet(temp, type)));
             }
             sw->value = nullptr;
             sw->finalize();
           } else {
-            assert(type == unreachable);
+            assert(type == Type::unreachable);
             // we don't need the br at all
             replaceCurrent(sw->value);
           }
         }
       }
     }
+    // TODO Handle br_on_exn
+
     // continue for general handling of everything, control flow or otherwise
     curr = getCurrent(); // we may have replaced it
     // we have changed children
     ReFinalizeNode().visit(curr);
-    // move everything to the prelude, if we need to: anything but constants
-    if (!curr->is<Const>()) {
-      if (curr->type == unreachable) {
-        ourPreludes.push_back(curr);
-        replaceCurrent(builder.makeUnreachable());
-      } else if (curr->type == none) {
-        if (!curr->is<Nop>()) {
-          ourPreludes.push_back(curr);
-          replaceCurrent(builder.makeNop());
-        }
-      } else {
-        // use a local
-        auto type = curr->type;
-        Index temp = builder.addVar(getFunction(), type);
-        ourPreludes.push_back(builder.makeSetLocal(temp, curr));
-        replaceCurrent(builder.makeGetLocal(temp, type));
-      }
+    if (curr->type == Type::unreachable) {
+      ourPreludes.push_back(curr);
+      replaceCurrent(builder.makeUnreachable());
+    } else if (curr->type.isConcrete()) {
+      // use a local
+      auto type = curr->type;
+      Index temp = builder.addVar(getFunction(), type);
+      ourPreludes.push_back(builder.makeLocalSet(temp, curr));
+      replaceCurrent(builder.makeLocalGet(temp, type));
     }
+
     // next, finish up: migrate our preludes if we can
     if (!ourPreludes.empty()) {
       auto* parent = getParent();
-      if (parent && !Flat::isControlFlowStructure(parent)) {
+      if (parent && !Properties::isControlFlowStructure(parent)) {
         auto& parentPreludes = preludes[parent];
         for (auto* prelude : ourPreludes) {
           parentPreludes.push_back(prelude);
@@ -259,7 +310,7 @@ struct Flatten : public WalkerPass<ExpressionStackWalker<Flatten, UnifiedExpress
   void visitFunction(Function* curr) {
     auto* originalBody = curr->body;
     // if the body is a block with a result, turn that into a return
-    if (isConcreteType(curr->body->type)) {
+    if (curr->body->type.isConcrete()) {
       curr->body = Builder(*getModule()).makeReturn(curr->body);
     }
     // the body may have preludes
@@ -275,9 +326,12 @@ private:
 
   // gets an expression, either by itself, or in a block with some
   // preludes (which we use up) for another expression before it
-  Expression* getPreludesWithExpression(Expression* preluder, Expression* after) {
+  Expression* getPreludesWithExpression(Expression* preluder,
+                                        Expression* after) {
     auto iter = preludes.find(preluder);
-    if (iter == preludes.end()) return after;
+    if (iter == preludes.end()) {
+      return after;
+    }
     // we have preludes
     auto& thePreludes = iter->second;
     auto* ret = Builder(*getModule()).makeBlock(thePreludes);
@@ -294,14 +348,12 @@ private:
     if (iter != breakTemps.end()) {
       return iter->second;
     } else {
-      return breakTemps[name] = Builder(*getModule()).addVar(getFunction(), type);
+      return breakTemps[name] =
+               Builder(*getModule()).addVar(getFunction(), type);
     }
   }
 };
 
-Pass *createFlattenPass() {
-  return new Flatten();
-}
+Pass* createFlattenPass() { return new Flatten(); }
 
 } // namespace wasm
-
