@@ -72,21 +72,23 @@
 // single outside block.
 //
 
-#include <wasm.h>
-#include <pass.h>
-#include <wasm-builder.h>
 #include <ir/branch-utils.h>
 #include <ir/effects.h>
 #include <ir/utils.h>
+#include <pass.h>
+#include <wasm-builder.h>
+#include <wasm.h>
 
 namespace wasm {
 
 // Looks for reasons we can't remove the values from breaks to an origin
-// For example, if there is a switch targeting us, we can't do it - we can't remove the value from other targets
+// For example, if there is a switch targeting us, we can't do it - we can't
+// remove the value from other targets
 struct ProblemFinder : public ControlFlowWalker<ProblemFinder> {
   Name origin;
   bool foundProblem = false;
-  // count br_ifs, and dropped br_ifs. if they don't match, then a br_if flow value is used, and we can't drop it
+  // count br_ifs, and dropped br_ifs. if they don't match, then a br_if flow
+  // value is used, and we can't drop it
   Index brIfs = 0;
   Index droppedBrIfs = 0;
   PassOptions& passOptions;
@@ -99,7 +101,8 @@ struct ProblemFinder : public ControlFlowWalker<ProblemFinder> {
         brIfs++;
       }
       // if the value has side effects, we can't remove it
-      if (EffectAnalyzer(passOptions, curr->value).hasSideEffects()) {
+      if (EffectAnalyzer(passOptions, getModule()->features, curr->value)
+            .hasSideEffects()) {
         foundProblem = true;
       }
     }
@@ -126,6 +129,11 @@ struct ProblemFinder : public ControlFlowWalker<ProblemFinder> {
     }
   }
 
+  void visitBrOnExn(BrOnExn* curr) {
+    // We should not take exnref value out of br_on_exn
+    foundProblem = true;
+  }
+
   bool found() {
     assert(brIfs >= droppedBrIfs);
     return foundProblem || brIfs > droppedBrIfs;
@@ -137,8 +145,11 @@ struct ProblemFinder : public ControlFlowWalker<ProblemFinder> {
 struct BreakValueDropper : public ControlFlowWalker<BreakValueDropper> {
   Name origin;
   PassOptions& passOptions;
+  BranchUtils::BranchSeekerCache& branchInfo;
 
-  BreakValueDropper(PassOptions& passOptions) : passOptions(passOptions) {}
+  BreakValueDropper(PassOptions& passOptions,
+                    BranchUtils::BranchSeekerCache& branchInfo)
+    : passOptions(passOptions), branchInfo(branchInfo) {}
 
   void visitBlock(Block* curr);
 
@@ -146,7 +157,7 @@ struct BreakValueDropper : public ControlFlowWalker<BreakValueDropper> {
     if (curr->value && curr->name == origin) {
       Builder builder(*getModule());
       auto* value = curr->value;
-      if (value->type == unreachable) {
+      if (value->type == Type::unreachable) {
         // the break isn't even reached
         replaceCurrent(value);
         return;
@@ -158,9 +169,11 @@ struct BreakValueDropper : public ControlFlowWalker<BreakValueDropper> {
   }
 
   void visitDrop(Drop* curr) {
-    // if we dropped a br_if whose value we removed, then we are now dropping a (block (drop value) (br_if)) with type none, which does not need a drop
-    // likewise, unreachable does not need to be dropped, so we just leave drops of concrete values
-    if (!isConcreteType(curr->value->type)) {
+    // if we dropped a br_if whose value we removed, then we are now dropping a
+    // (block (drop value) (br_if)) with type none, which does not need a drop
+    // likewise, unreachable does not need to be dropped, so we just leave drops
+    // of concrete values
+    if (!curr->value->type.isConcrete()) {
       replaceCurrent(curr->value);
     }
   }
@@ -168,7 +181,19 @@ struct BreakValueDropper : public ControlFlowWalker<BreakValueDropper> {
 
 static bool hasUnreachableChild(Block* block) {
   for (auto* test : block->list) {
-    if (test->type == unreachable) {
+    if (test->type == Type::unreachable) {
+      return true;
+    }
+  }
+  return false;
+}
+
+// Checks for code after an unreachable element.
+static bool hasDeadCode(Block* block) {
+  auto& list = block->list;
+  auto size = list.size();
+  for (size_t i = 1; i < size; i++) {
+    if (list[i - 1]->type == Type::unreachable) {
       return true;
     }
   }
@@ -176,10 +201,14 @@ static bool hasUnreachableChild(Block* block) {
 }
 
 // core block optimizer routine
-static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions) {
+static void optimizeBlock(Block* curr,
+                          Module* module,
+                          PassOptions& passOptions,
+                          BranchUtils::BranchSeekerCache& branchInfo) {
+  auto& list = curr->list;
+  // Main merging loop.
   bool more = true;
   bool changed = false;
-  auto& list = curr->list;
   while (more) {
     more = false;
     for (size_t i = 0; i < list.size(); i++) {
@@ -192,7 +221,8 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
       Loop* loop = nullptr;
       // To to handle a non-block child.
       if (!childBlock) {
-        // if we have a child that is (drop (block ..)) then we can move the drop into the block, and remove br values. this allows more merging,
+        // if we have a child that is (drop (block ..)) then we can move the
+        // drop into the block, and remove br values. this allows more merging,
         if (auto* drop = list[i]->dynCast<Drop>()) {
           childBlock = drop->value->dynCast<Block>();
           if (childBlock) {
@@ -205,13 +235,14 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
               Expression* expression = childBlock;
               // check if it's ok to remove the value from all breaks to us
               ProblemFinder finder(passOptions);
+              finder.setModule(module);
               finder.origin = childBlock->name;
               finder.walk(expression);
               if (finder.found()) {
                 childBlock = nullptr;
               } else {
                 // fix up breaks
-                BreakValueDropper fixer(passOptions);
+                BreakValueDropper fixer(passOptions, branchInfo);
                 fixer.origin = childBlock->name;
                 fixer.setModule(module);
                 fixer.walk(expression);
@@ -221,7 +252,7 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
               // we can do it!
               // reuse the drop, if we still need it
               auto* last = childBlock->list.back();
-              if (isConcreteType(last->type)) {
+              if (last->type.isConcrete()) {
                 drop->value = last;
                 drop->finalize();
                 childBlock->list.back() = drop;
@@ -239,52 +270,110 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
           // TODO: handle (loop (loop - the bodies of loops may not be blocks
         }
       }
-      // If no block, or a block that might be broken out of, we can't do anything.
-      if (!childBlock || childBlock->name.is()) continue;
+      // If no block, we can't do anything.
+      if (!childBlock) {
+        continue;
+      }
       auto& childList = childBlock->list;
       auto childSize = childList.size();
-      if (childSize == 0) continue;
-      // If this is a loop, we may be removing only the tail.
-      Index start = 0;
-      Index end = childSize;
+      if (childSize == 0) {
+        continue;
+      }
+      // If the child has items after an unreachable, ignore it - dce should
+      // have been run, and we prefer to not handle the complexity here.
+      if (hasDeadCode(childBlock)) {
+        continue;
+      }
+      // In some cases we can remove only the head or the tail of the block,
+      // and must keep some things in the child block.
+      Index keepStart = childSize;
+      Index keepEnd = 0;
+      // For a block with a name, we may only be able to remove a head, up
+      // to the first item that branches to the block.
+      if (childBlock->name.is()) {
+        // If it has a concrete value, then breaks may be sending it a value,
+        // and we'd need to handle that. TODO
+        if (childBlock->type.isConcrete()) {
+          continue;
+        }
+        auto childName = childBlock->name;
+        for (size_t j = 0; j < childSize; j++) {
+          auto* item = childList[j];
+          if (branchInfo.hasBranch(item, childName)) {
+            // We can't remove this from the child.
+            keepStart = j;
+            keepEnd = childSize;
+            break;
+          }
+        }
+      }
+      // For a loop, we may only be able to remove a tail
       if (loop) {
         auto childName = loop->name;
         for (auto j = int(childSize - 1); j >= 0; j--) {
           auto* item = childList[j];
-          if (BranchUtils::BranchSeeker::hasNamed(item, childName)) {
+          if (BranchUtils::BranchSeeker::has(item, childName)) {
             // We can't remove this from the child.
-            start = j + 1;
+            keepStart = 0;
+            keepEnd = std::max(Index(j + 1), keepEnd);
             break;
           }
         }
-        // Maybe there's nothing to do.
-        if (start == end) continue;
-        // If we can only do part of the block, and if the block has a flowing value, we
-        // would need special handling for that - not worth it, probably TODO
-        if (start > 0 && isConcreteType(childList.back()->type)) {
+        // If we can only do part of the block, and if the block has a flowing
+        // value, we would need special handling for that - not worth it,
+        // probably TODO
+        // FIXME is this not handled by the drop later down?
+        if (keepEnd < childSize && childList.back()->type.isConcrete()) {
           continue;
         }
       }
+      // Maybe there's nothing to do, if we must keep it all in the
+      // child anyhow.
+      if (keepStart == 0 && keepEnd == childSize) {
+        continue;
+      }
+      // There is something to do!
+      bool keepingPart = keepStart < keepEnd;
+      // Create a new merged list, and fill in the code before the
+      // child block we are merging in. TODO better efficiency
       ExpressionList merged(module->allocator);
       for (size_t j = 0; j < i; j++) {
         merged.push_back(list[j]);
       }
-      // If we can't merge it all, keep the child.
-      if (start > 0) {
-        merged.push_back(child);
-      }
-      for (Index j = start; j < end; j++) {
+      // Add the head of the block - the things at the start we do
+      // not need to keep.
+      for (Index j = 0; j < keepStart; j++) {
         merged.push_back(childList[j]);
       }
-      // If we didn't merge it all, update the child.
-      if (start > 0) {
-        childList.resize(start);
+      // If we can't merge it all, keep and filter the child.
+      if (keepingPart) {
+        merged.push_back(child);
+        // Filter the child.
+        ExpressionList filtered(module->allocator);
+        for (Index j = keepStart; j < keepEnd; j++) {
+          filtered.push_back(childList[j]);
+        }
+        // Add the tail of the block - the things at the end we do
+        // not need to keep.
+        for (Index j = keepEnd; j < childSize; j++) {
+          merged.push_back(childList[j]);
+        }
+        // Update the child.
+        childList.swap(filtered);
         // We may have removed unreachable items.
         childBlock->finalize();
-        if (auto* loop = child->dynCast<Loop>()) {
+        if (loop) {
           loop->finalize();
         }
+        // Note that we modify the child block here, which invalidates info
+        // in branchInfo. However, as we have scanned the parent, we have
+        // already forgotten the child's info, so there is nothing to do here
+        // for the child.
+        // (We also don't need to do anything for the parent - we move code
+        // from a child into the parent, but that doesn't change the total
+        // branches in the parent.)
       }
+      // Add the rest of the parent block after the child.
       for (size_t j = i + 1; j < list.size(); j++) {
         merged.push_back(list[j]);
       }
@@ -292,7 +381,7 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
       if (!merged.empty()) {
         auto* last = merged.back();
         for (auto*& item : merged) {
-          if (item != last && isConcreteType(item->type)) {
+          if (item != last && item->type.isConcrete()) {
             Builder builder(*module);
             item = builder.makeDrop(item);
           }
@@ -304,11 +393,13 @@ static void optimizeBlock(Block* curr, Module* module, PassOptions& passOptions)
       break;
     }
   }
-  if (changed) curr->finalize(curr->type);
+  if (changed) {
+    curr->finalize(curr->type);
+  }
 }
 
 void BreakValueDropper::visitBlock(Block* curr) {
-  optimizeBlock(curr, getModule(), passOptions);
+  optimizeBlock(curr, getModule(), passOptions, branchInfo);
 }
 
 struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
@@ -316,8 +407,10 @@ struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
 
   Pass* create() override { return new MergeBlocks; }
 
-  void visitBlock(Block *curr) {
-    optimizeBlock(curr, getModule(), getPassOptions());
+  BranchUtils::BranchSeekerCache branchInfo;
+
+  void visitBlock(Block* curr) {
+    optimizeBlock(curr, getModule(), getPassOptions(), branchInfo);
   }
 
   // given
@@ -336,31 +429,49 @@ struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
   //   (..other..children..)
   //  )
   // )
-  // at which point the block is on the outside and potentially mergeable with an outer block
-  Block* optimize(Expression* curr, Expression*& child, Block* outer = nullptr, Expression** dependency1 = nullptr, Expression** dependency2 = nullptr) {
-    if (!child) return outer;
+  // at which point the block is on the outside and potentially mergeable with
+  // an outer block
+  Block* optimize(Expression* curr,
+                  Expression*& child,
+                  Block* outer = nullptr,
+                  Expression** dependency1 = nullptr,
+                  Expression** dependency2 = nullptr) {
+    if (!child) {
+      return outer;
+    }
+    FeatureSet features = getModule()->features;
     if ((dependency1 && *dependency1) || (dependency2 && *dependency2)) {
-      // there are dependencies, things we must be reordered through. make sure no problems there
-      EffectAnalyzer childEffects(getPassOptions(), child);
-      if (dependency1 && *dependency1 && EffectAnalyzer(getPassOptions(), *dependency1).invalidates(childEffects)) return outer;
-      if (dependency2 && *dependency2 && EffectAnalyzer(getPassOptions(), *dependency2).invalidates(childEffects)) return outer;
+      // there are dependencies, things we must be reordered through. make sure
+      // no problems there
+      EffectAnalyzer childEffects(getPassOptions(), features, child);
+      if (dependency1 && *dependency1 &&
+          EffectAnalyzer(getPassOptions(), features, *dependency1)
+            .invalidates(childEffects)) {
+        return outer;
+      }
+      if (dependency2 && *dependency2 &&
+          EffectAnalyzer(getPassOptions(), features, *dependency2)
+            .invalidates(childEffects)) {
+        return outer;
+      }
     }
     if (auto* block = child->dynCast<Block>()) {
       if (!block->name.is() && block->list.size() >= 2) {
-        // if we move around unreachable code, type changes could occur. avoid that, as
-        // anyhow it means we should have run dce before getting here
-        if (curr->type == none && hasUnreachableChild(block)) {
-          // moving the block to the outside would replace a none with an unreachable
+        // if we move around unreachable code, type changes could occur. avoid
+        // that, as anyhow it means we should have run dce before getting here
+        if (curr->type == Type::none && hasUnreachableChild(block)) {
+          // moving the block to the outside would replace a none with an
+          // unreachable
           return outer;
         }
         auto* back = block->list.back();
-        if (back->type == unreachable) {
-          // curr is not reachable, dce could remove it; don't try anything fancy
-          // here
+        if (back->type == Type::unreachable) {
+          // curr is not reachable, dce could remove it; don't try anything
+          // fancy here
           return outer;
         }
-        // we are going to replace the block with the final element, so they should
-        // be identically typed
+        // we are going to replace the block with the final element, so they
+        // should be identically typed
         if (block->type != back->type) {
           return outer;
         }
@@ -386,18 +497,10 @@ struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
     return outer;
   }
 
-  void visitUnary(Unary* curr) {
-    optimize(curr, curr->value);
-  }
-  void visitSetLocal(SetLocal* curr) {
-    optimize(curr, curr->value);
-  }
-  void visitLoad(Load* curr) {
-    optimize(curr, curr->ptr);
-  }
-  void visitReturn(Return* curr) {
-    optimize(curr, curr->value);
-  }
+  void visitUnary(Unary* curr) { optimize(curr, curr->value); }
+  void visitLocalSet(LocalSet* curr) { optimize(curr, curr->value); }
+  void visitLoad(Load* curr) { optimize(curr, curr->ptr); }
+  void visitReturn(Return* curr) { optimize(curr, curr->value); }
 
   void visitBinary(Binary* curr) {
     optimize(curr, curr->right, optimize(curr, curr->left), &curr->left);
@@ -409,15 +512,24 @@ struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
     optimize(curr, curr->value, optimize(curr, curr->ptr), &curr->ptr);
   }
   void optimizeTernary(Expression* curr,
-		       Expression*& first, Expression*& second, Expression*& third) {
+                       Expression*& first,
+                       Expression*& second,
+                       Expression*& third) {
     // TODO: for now, just stop when we see any side effect. instead, we could
     //       check effects carefully for reordering
+    FeatureSet features = getModule()->features;
     Block* outer = nullptr;
-    if (EffectAnalyzer(getPassOptions(), first).hasSideEffects()) return;
+    if (EffectAnalyzer(getPassOptions(), features, first).hasSideEffects()) {
+      return;
+    }
     outer = optimize(curr, first, outer);
-    if (EffectAnalyzer(getPassOptions(), second).hasSideEffects()) return;
+    if (EffectAnalyzer(getPassOptions(), features, second).hasSideEffects()) {
+      return;
+    }
     outer = optimize(curr, second, outer);
-    if (EffectAnalyzer(getPassOptions(), third).hasSideEffects()) return;
+    if (EffectAnalyzer(getPassOptions(), features, third).hasSideEffects()) {
+      return;
+    }
     optimize(curr, third, outer);
   }
   void visitAtomicCmpxchg(AtomicCmpxchg* curr) {
@@ -428,44 +540,64 @@ struct MergeBlocks : public WalkerPass<PostWalker<MergeBlocks>> {
     optimizeTernary(curr, curr->ifTrue, curr->ifFalse, curr->condition);
   }
 
-  void visitDrop(Drop* curr) {
-    optimize(curr, curr->value);
-  }
+  void visitDrop(Drop* curr) { optimize(curr, curr->value); }
 
   void visitBreak(Break* curr) {
     optimize(curr, curr->condition, optimize(curr, curr->value), &curr->value);
   }
+
   void visitSwitch(Switch* curr) {
     optimize(curr, curr->condition, optimize(curr, curr->value), &curr->value);
   }
 
-  template<typename T>
-  void handleCall(T* curr) {
+  template<typename T> void handleCall(T* curr) {
     Block* outer = nullptr;
     for (Index i = 0; i < curr->operands.size(); i++) {
-      if (EffectAnalyzer(getPassOptions(), curr->operands[i]).hasSideEffects()) return;
+      if (EffectAnalyzer(
+            getPassOptions(), getModule()->features, curr->operands[i])
+            .hasSideEffects()) {
+        return;
+      }
       outer = optimize(curr, curr->operands[i], outer);
     }
-    return;
   }
 
-  void visitCall(Call* curr) {
-    handleCall(curr);
-  }
+  void visitCall(Call* curr) { handleCall(curr); }
 
   void visitCallIndirect(CallIndirect* curr) {
+    FeatureSet features = getModule()->features;
     Block* outer = nullptr;
     for (Index i = 0; i < curr->operands.size(); i++) {
-      if (EffectAnalyzer(getPassOptions(), curr->operands[i]).hasSideEffects()) return;
+      if (EffectAnalyzer(getPassOptions(), features, curr->operands[i])
+            .hasSideEffects()) {
+        return;
+      }
       outer = optimize(curr, curr->operands[i], outer);
     }
-    if (EffectAnalyzer(getPassOptions(), curr->target).hasSideEffects()) return;
+    if (EffectAnalyzer(getPassOptions(), features, curr->target)
+          .hasSideEffects()) {
+      return;
+    }
     optimize(curr, curr->target, outer);
   }
+
+  void visitThrow(Throw* curr) {
+    Block* outer = nullptr;
+    for (Index i = 0; i < curr->operands.size(); i++) {
+      if (EffectAnalyzer(
+            getPassOptions(), getModule()->features, curr->operands[i])
+            .hasSideEffects()) {
+        return;
+      }
+      outer = optimize(curr, curr->operands[i], outer);
+    }
+  }
+
+  void visitRethrow(Rethrow* curr) { optimize(curr, curr->exnref); }
+
+  void visitBrOnExn(BrOnExn* curr) { optimize(curr, curr->exnref); }
 };
 
-Pass *createMergeBlocksPass() {
-  return new MergeBlocks();
-}
+Pass* createMergeBlocksPass() { return new MergeBlocks(); }
 
 } // namespace wasm
