@@ -18,46 +18,79 @@
 // Shared execution result checking code
 //
 
-#include "wasm.h"
 #include "shell-interface.h"
-#include "ir/import-utils.h"
+#include "wasm.h"
 
 namespace wasm {
 
+typedef std::vector<Literal> Loggings;
+
+// Logs every relevant import call parameter.
+struct LoggingExternalInterface : public ShellExternalInterface {
+  Loggings& loggings;
+
+  LoggingExternalInterface(Loggings& loggings) : loggings(loggings) {}
+
+  Literals callImport(Function* import, LiteralList& arguments) override {
+    if (import->module == "fuzzing-support") {
+      std::cout << "[LoggingExternalInterface logging";
+      loggings.push_back(Literal()); // buffer with a None between calls
+      for (auto argument : arguments) {
+        std::cout << ' ' << argument;
+        loggings.push_back(argument);
+      }
+      std::cout << "]\n";
+    }
+    return {};
+  }
+};
+
 // gets execution results from a wasm module. this is useful for fuzzing
 //
 // we can only get results when there are no imports. we then call each method
 // that has a result, with some values
 struct ExecutionResults {
-  std::map<Name, Literal> results;
+  std::map<Name, Literals> results;
+  Loggings loggings;
 
   // get results of execution
   void get(Module& wasm) {
-    if (ImportInfo(wasm).getNumImports() > 0) {
-      std::cout << "[fuzz-exec] imports, so quitting\n";
-      return;
-    }
-    ShellExternalInterface interface;
+    LoggingExternalInterface interface(loggings);
     try {
       ModuleInstance instance(wasm, &interface);
-      // execute all exported methods (that are therefore preserved through opts)
+      // execute all exported methods (that are therefore preserved through
+      // opts)
       for (auto& exp : wasm.exports) {
-        if (exp->kind != ExternalKind::Function) continue;
+        if (exp->kind != ExternalKind::Function) {
+          continue;
+        }
+        std::cout << "[fuzz-exec] calling " << exp->name << "\n";
         auto* func = wasm.getFunction(exp->value);
-        if (func->result != none) {
+        if (func->sig.results != Type::none) {
           // this has a result
-          results[exp->name] = run(func, wasm, instance);
-          std::cout << "[fuzz-exec] note result: " << exp->name << " => " << results[exp->name] << '\n';
+          Literals ret = run(func, wasm, instance);
+          // We cannot compare funcrefs by name because function names can
+          // change (after duplicate function elimination or roundtripping)
+          // while the function contents are still the same
+          for (Literal& val : ret) {
+            if (val.type == Type::funcref && !val.isNull()) {
+              val = Literal::makeFunc(Name("funcref"));
+            }
+          }
+          results[exp->name] = ret;
+          // ignore the result if we hit an unreachable and returned no value
+          if (ret.size() > 0) {
+            std::cout << "[fuzz-exec] note result: " << exp->name << " => "
+                      << ret << '\n';
+          }
         } else {
           // no result, run it anyhow (it might modify memory etc.)
           run(func, wasm, instance);
-          std::cout << "[fuzz-exec] no result for void func: " << exp->name << '\n';
         }
       }
     } catch (const TrapException&) {
       // may throw in instance creation (init of offsets)
     }
-    std::cout << "[fuzz-exec] " << results.size() << " results noted\n";
   }
 
   // get current results and check them against previous ones
@@ -66,9 +99,8 @@ struct ExecutionResults {
     optimizedResults.get(wasm);
     if (optimizedResults != *this) {
       std::cout << "[fuzz-exec] optimization passes changed execution results";
-      abort();
+      exit(1);
     }
-    std::cout << "[fuzz-exec] " << results.size() << " results match\n";
   }
 
   bool operator==(ExecutionResults& other) {
@@ -76,50 +108,52 @@ struct ExecutionResults {
       auto name = iter.first;
       if (results.find(name) == results.end()) {
         std::cout << "[fuzz-exec] missing " << name << '\n';
-        abort();
+        return false;
       }
       std::cout << "[fuzz-exec] comparing " << name << '\n';
       if (results[name] != other.results[name]) {
-        std::cout << "not identical!\n";
-        abort();
+        std::cout << "not identical! " << results[name]
+                  << " != " << other.results[name] << "\n";
+        return false;
       }
     }
+    if (loggings != other.loggings) {
+      std::cout << "logging not identical!\n";
+      return false;
+    }
     return true;
   }
 
-  bool operator!=(ExecutionResults& other) {
-    return !((*this) == other);
-  }
+  bool operator!=(ExecutionResults& other) { return !((*this) == other); }
 
-  Literal run(Function* func, Module& wasm) {
-    ShellExternalInterface interface;
+  Literals run(Function* func, Module& wasm) {
+    LoggingExternalInterface interface(loggings);
     try {
       ModuleInstance instance(wasm, &interface);
       return run(func, wasm, instance);
     } catch (const TrapException&) {
       // may throw in instance creation (init of offsets)
-      return Literal();
+      return {};
     }
   }
 
-  Literal run(Function* func, Module& wasm, ModuleInstance& instance) {
+  Literals run(Function* func, Module& wasm, ModuleInstance& instance) {
     try {
       LiteralList arguments;
       // init hang support, if present
-      if (wasm.getFunctionOrNull("hangLimitInitializer")) {
-        instance.callFunction("hangLimitInitializer", arguments);
+      if (auto* ex = wasm.getExportOrNull("hangLimitInitializer")) {
+        instance.callFunction(ex->value, arguments);
       }
       // call the method
-      for (Type param : func->params) {
+      for (const auto& param : func->sig.params) {
         // zeros in arguments TODO: more?
-        arguments.push_back(Literal(param));
+        arguments.push_back(Literal::makeZero(param));
       }
       return instance.callFunction(func->name, arguments);
     } catch (const TrapException&) {
-      return Literal();
+      return {};
     }
   }
 };
 
 } // namespace wasm
-
