@@ -25,28 +25,32 @@
 //
 
 #include "wasm-io.h"
-#include "wasm-s-parser.h"
+#include "support/debug.h"
 #include "wasm-binary.h"
+#include "wasm-s-parser.h"
 
 namespace wasm {
 
-static void readTextData(std::string& input, Module& wasm) {
+#define DEBUG_TYPE "writer"
+
+static void readTextData(std::string& input, Module& wasm, IRProfile profile) {
   SExpressionParser parser(const_cast<char*>(input.c_str()));
   Element& root = *parser.root;
-  SExpressionWasmBuilder builder(wasm, *root[0]);
-
+  SExpressionWasmBuilder builder(wasm, *root[0], profile);
 }
 
 void ModuleReader::readText(std::string filename, Module& wasm) {
-  if (debug) std::cerr << "reading text from " << filename << "\n";
-  auto input(read_file<std::string>(filename, Flags::Text, debug ? Flags::Debug : Flags::Release));
-  readTextData(input, wasm);
+  BYN_TRACE("reading text from " << filename << "\n");
+  auto input(read_file<std::string>(filename, Flags::Text));
+  readTextData(input, wasm, profile);
 }
 
-static void readBinaryData(std::vector<char>& input, Module& wasm,
-                           std::string sourceMapFilename, bool debug) {
+void ModuleReader::readBinaryData(std::vector<char>& input,
+                                  Module& wasm,
+                                  std::string sourceMapFilename) {
   std::unique_ptr<std::ifstream> sourceMapStream;
-  WasmBinaryBuilder parser(wasm, input, debug);
+  WasmBinaryBuilder parser(wasm, input);
+  parser.setDWARF(DWARF);
   if (sourceMapFilename.size()) {
     sourceMapStream = make_unique<std::ifstream>();
     sourceMapStream->open(sourceMapFilename);
@@ -58,24 +62,27 @@ static void readBinaryData(std::vector<char>& input, Module& wasm,
   }
 }
 
-void ModuleReader::readBinary(std::string filename, Module& wasm,
+void ModuleReader::readBinary(std::string filename,
+                              Module& wasm,
                               std::string sourceMapFilename) {
-  if (debug) std::cerr << "reading binary from " << filename << "\n";
-  auto input(read_file<std::vector<char>>(filename, Flags::Binary, debug ? Flags::Debug : Flags::Release));
-  readBinaryData(input, wasm, sourceMapFilename, debug);
+  BYN_TRACE("reading binary from " << filename << "\n");
+  auto input(read_file<std::vector<char>>(filename, Flags::Binary));
+  readBinaryData(input, wasm, sourceMapFilename);
 }
 
 bool ModuleReader::isBinaryFile(std::string filename) {
   std::ifstream infile;
   std::ios_base::openmode flags = std::ifstream::in | std::ifstream::binary;
   infile.open(filename, flags);
-  char buffer[4] = { 1, 2, 3, 4 };
+  char buffer[4] = {1, 2, 3, 4};
   infile.read(buffer, 4);
   infile.close();
-  return buffer[0] == '\0' && buffer[1] == 'a' && buffer[2] == 's' && buffer[3] == 'm';
+  return buffer[0] == '\0' && buffer[1] == 'a' && buffer[2] == 's' &&
+         buffer[3] == 'm';
 }
 
-void ModuleReader::read(std::string filename, Module& wasm,
+void ModuleReader::read(std::string filename,
+                        Module& wasm,
                         std::string sourceMapFilename) {
   // empty filename means read from stdin
   if (!filename.size()) {
@@ -87,7 +94,8 @@ void ModuleReader::read(std::string filename, Module& wasm,
   } else {
     // default to text
     if (sourceMapFilename.size()) {
-      std::cerr << "Binaryen ModuleReader::read() - source map filename provided, but file appears to not be binary\n";
+      std::cerr << "Binaryen ModuleReader::read() - source map filename "
+                   "provided, but file appears to not be binary\n";
     }
     readText(filename, wasm);
   }
@@ -96,33 +104,35 @@ void ModuleReader::read(std::string filename, Module& wasm,
 // TODO: reading into a vector<char> then copying into a string is unnecessarily
 // inefficient. It would be better to read just once into a stringstream.
 void ModuleReader::readStdin(Module& wasm, std::string sourceMapFilename) {
-  std::vector<char> input = read_stdin(debug ? Flags::Debug : Flags::Release);
+  std::vector<char> input = read_stdin();
   if (input.size() >= 4 && input[0] == '\0' && input[1] == 'a' &&
       input[2] == 's' && input[3] == 'm') {
-    readBinaryData(input, wasm, sourceMapFilename, debug);
+    readBinaryData(input, wasm, sourceMapFilename);
   } else {
     std::ostringstream s;
     s.write(input.data(), input.size());
     s << '\0';
     std::string input_str = s.str();
-    readTextData(input_str, wasm);
+    readTextData(input_str, wasm, profile);
   }
 }
 
+#undef DEBUG_TYPE
+#define DEBUG_TYPE "writer"
 
 void ModuleWriter::writeText(Module& wasm, Output& output) {
   WasmPrinter::printModule(&wasm, output.getStream());
 }
 
 void ModuleWriter::writeText(Module& wasm, std::string filename) {
-  if (debug) std::cerr << "writing text to " << filename << "\n";
-  Output output(filename, Flags::Text, debug ? Flags::Debug : Flags::Release);
+  BYN_TRACE("writing text to " << filename << "\n");
+  Output output(filename, Flags::Text);
   writeText(wasm, output);
 }
 
 void ModuleWriter::writeBinary(Module& wasm, Output& output) {
-  BufferWithRandomAccess buffer(debug);
-  WasmBinaryWriter writer(&wasm, buffer, debug);
+  BufferWithRandomAccess buffer;
+  WasmBinaryWriter writer(&wasm, buffer);
   // if debug info is used, then we want to emit the names section
   writer.setNamesSection(debugInfo);
   std::unique_ptr<std::ofstream> sourceMapStream;
@@ -131,7 +141,9 @@ void ModuleWriter::writeBinary(Module& wasm, Output& output) {
     sourceMapStream->open(sourceMapFilename);
     writer.setSourceMap(sourceMapStream.get(), sourceMapUrl);
   }
-  if (symbolMap.size() > 0) writer.setSymbolMap(symbolMap);
+  if (symbolMap.size() > 0) {
+    writer.setSymbolMap(symbolMap);
+  }
   writer.write();
   buffer.writeTo(output);
   if (sourceMapStream) {
@@ -140,8 +152,8 @@ void ModuleWriter::writeBinary(Module& wasm, Output& output) {
 }
 
 void ModuleWriter::writeBinary(Module& wasm, std::string filename) {
-  if (debug) std::cerr << "writing binary to " << filename << "\n";
-  Output output(filename, Flags::Binary, debug ? Flags::Debug : Flags::Release);
+  BYN_TRACE("writing binary to " << filename << "\n");
+  Output output(filename, Flags::Binary);
   writeBinary(wasm, output);
 }
 
@@ -161,4 +173,4 @@ void ModuleWriter::write(Module& wasm, std::string filename) {
   }
 }
 
-}
+} // namespace wasm
